
#if __LINUX_X86__
	#include <stdio.h>
	#include <unistd.h>
	#include <sys/uio.h>
	#include <sys/ioctl.h>
	#include <sys/time.h>
	#include <errno.h>
	#include <pthread.h>
	#include <fcntl.h>
#endif

#include "XC_Engine.h"
#include "UnLinker.h"


class FSurfaceInfo;
class FTransTexture;
class FSurfaceFacet;
#include "UnRenDev.h" //Avoid including UnRender.h to reduce compile times
#include "XC_Networking.h"
#include "UnCon.h"
#include "UnXC_Lev.h"
#include "UnXC_Col.h"
#include "UnXC_Travel.h"
#include "XC_LZMA.h"

IMPLEMENT_CLASS(UXC_Level);


struct FBatch
{
	FBatch* NextBatch;
	INT Count;
	INT Size;
	INT Indices[4096]; //Size is variable

	FBatch( FBatch* InNext=NULL)
		: NextBatch(InNext)
		, Count(0)
	{}
	
	static FBatch* AllocateBatch( INT BatchSize, FBatch* Other=NULL)
	{
		guard(FBatch::AllocateBatch);

		BatchSize = Clamp( BatchSize, 32, 4096);
		FBatch* Result = (FBatch*)appMalloc( 12 + BatchSize * 4, TEXT("FBatch"));
		Result->NextBatch = Other;
		Result->Count = 0;
		Result->Size = BatchSize;
		return Result;

		unguard;
	}
};

struct FNameBatchMap : public TMap<INT,FBatch*>
{
	TArray<INT> TransientNames;
	
	INT GetNameHash( const TCHAR* InStr)
	{
		guard(FNameBatchMap::GetNameHash);
		INT Len = Min( appStrlen(InStr), 63);
		TCHAR NewName[64];
		NewName[Len--] = 0;
		while ( (Len > 0) && appIsDigit(InStr[Len]) )
			NewName[Len--] = 0;
		for ( ; Len >= 0 ; Len-- )
			NewName[Len] = InStr[Len];
		return appStrihash( NewName) & 0x7FFFFFFF;
		unguard;
	}
	FName Request( UObject* Parent, UClass* ForClass )
	{
		guard(FNameBatchMap::Request);
		INT iHash = GetNameHash( ForClass->GetName() );
		if ( !Hash )
			Rehash();
		FBatch* Batch = FindRef( iHash);
		while ( Batch )
		{
			while ( Batch->Count > 0 )
			{
				//If object matches class and is deleted, pick it up
				Batch->Count--;
				FNameEntry* Entry = FName::GetEntry( Batch->Indices[Batch->Count] );
				if ( Entry && (Entry->Flags & RF_Transient) && (GetNameHash( Entry->Name) == iHash) && !UObject::StaticFindObject( NULL, Parent, Entry->Name ) )
					return *((FName*)&Entry->Index);
			}
			FBatch* NextBatch = Batch->NextBatch;
			delete Batch;
			Batch = NextBatch;
			if ( Batch )
			{
//				debugf( TEXT("Next batch in queue: (%i/%i)"), Batch->Count, Batch->Size);
				Set( iHash, Batch);
				check( FindRef(iHash) );
			}
			else
				Remove( iHash);
		}
		return NAME_None;
		unguard;
	}

	void Store( FName AName)
	{
		guard(FNameBatchMap::Store);
		INT iHash = GetNameHash( *AName);
		FBatch* Batch = FindRef( iHash );
		guard(BatchSetup);
		if ( !Batch )
		{
			Batch = FBatch::AllocateBatch(32);
			Set( iHash, Batch);
		}
		else if ( Batch->Count >= Batch->Size )
		{
			Batch = FBatch::AllocateBatch( Batch->Size + (Batch->Size >> 2) + (iHash%8), Batch);
			Set( iHash, Batch);
		}
		check( Batch->Count < Batch->Size );	
		unguard;
		//Add single name
		Batch->Indices[Batch->Count++] = *(INT*)&AName;
		unguard;
	}
	
	void Cleanup()
	{
		guard(FNameBatchMap::Cleanup);
		INT i;
		for ( i=0 ; i<TransientNames.Num() ; i++ )
		{
			FNameEntry* NEntry = FName::GetEntry(TransientNames(i));
			if ( NEntry )
				NEntry->Flags &= ~RF_Destroyed;
		}
		SafeEmpty(TransientNames);
/*		INT MaxNames = FName::GetMaxNames();
		for ( i=0 ; i<MaxNames ; i++ )
		{
			FNameEntry* NEntry = FName::GetEntry(i);
			if ( NEntry )
				NEntry->Flags &= ~RF_Destroyed;
		}
*/		for ( i=0 ; i<Pairs.Num() ; i++ )
		{
			FBatch* Batch = Pairs(i).Value;
			while ( Batch )
			{
				FBatch* Next = Batch->NextBatch;
				delete Batch;
				Batch = Next;
			}
		}
		Empty();
		unguard;
	}
};

static INT BatchData[8] = { 0, 0, 0, 0, 8, 0, 0, 0 };
static FNameBatchMap* UnusedNamePool = (FNameBatchMap*)BatchData;

//Stats
static INT ST_Traces = 0;

//Map loader stuff
static volatile INT GMapLoadCount = 0;
static volatile UBOOL GIsLoadingMap = 0; //Needed to avoid critical errors
static volatile UBOOL GLockPackageMap = 0;

//Throw warning to compiler if one of these hacks is missing
inline void CompilerCheck()
{
/** Add these extra bitfield values in AActor as follows:

    BITFIELD bClientDemoNetFunc:1;
	//
	BITFIELD bNotRelevantToOwner:1; //Lag compensators love this
	BITFIELD bNetDirty:1;
	//
    class UClass* RenderIteratorClass GCC_PACK(4);

*/
	AActor* TestActor;
	TestActor->bNotRelevantToOwner = TestActor->bNotRelevantToOwner;
	TestActor->bNetDirty = TestActor->bNetDirty;
}


enum EOwned { EPri_Owned };
enum EGeneral { EPri_General };


struct FActorPriority
{
	INT			Priority;	// Update priority, higher = more important.
	AActor*			Actor;		// Actor.
	UActorChannel*	Channel;	// Actor channel.
	UBOOL		SuperRelevant; //Skip owner checks later
	FActorPriority()
	{}
	//Use this on owned actors (NetOwner)
	FActorPriority( EOwned, FVector4& Location, FVector4& Dir, UNetConnection* InConnection, UActorChannel* InChannel, AActor* Target)
	{
		Actor = Target;
		Channel = InChannel;
		if ( Actor->bNotRelevantToOwner )
		{
			SuperRelevant = -1;
			Priority = -3000000;
			return;
		}
		FLOAT Time  = Channel ? (InConnection->Driver->Time - Channel->LastUpdateTime) : InConnection->Driver->SpawnPrioritySeconds;
		FLOAT Something = (Actor == InConnection->Actor) ? 4.f : 1.5f;
		SuperRelevant = 1;
		if ( !Actor->bAlwaysRelevant )
		{
			if ( InConnection->Actor->Weapon == Actor )
				Something += 1.0f;
			else if ( (Actor->RemoteRole > ROLE_DumbProxy) || (Actor->Physics != PHYS_None) )
			{
				FVector4 Delta( Actor->Location - Location);
				Something += (Delta.Dot4(Dir) >= 0) ? 0.5f : -0.5f;
			}
		}

		AActor* Sent = (Channel && Channel->Recent.Num()) ? (AActor*) &Channel->Recent(0) : NULL;
		FLOAT Pri = Actor->GetNetPriority( Sent, Time, InConnection->BestLag);
		Priority = appRound( (Something + 3.0f) * Pri * 65536.f);

		if ( Actor->bNetOptional )
			Priority -= 3000000;
	}
	//Don't use this on owned actors (NetOwner)
	FActorPriority( EGeneral, FVector4& Location, FVector4& Dir, UNetConnection* InConnection, UActorChannel* InChannel, AActor* Target)
	{
		Actor = Target;
		Channel = InChannel;
		FLOAT Time  = Channel ? (InConnection->Driver->Time - Channel->LastUpdateTime) : InConnection->Driver->SpawnPrioritySeconds;
		FLOAT Something = 0.f;
		SuperRelevant = 0;

		if ( !Actor->bAlwaysRelevant )
		{
			if ( (Actor->RemoteRole > ROLE_DumbProxy) || (Actor->Physics != PHYS_None) )
			{
				FVector4 Delta( Actor->Location - Location);
				Something = (Delta.Dot4(Dir) >= 0) ? 0.5f : -0.5f;
			}
		}
		APlayerPawn* NetOwner = InConnection->Actor;
		AActor* Sent = (Channel && Channel->Recent.Num()) ? (AActor*) &Channel->Recent(0) : NULL;

		if ( Actor == NetOwner->ViewTarget )
		{
			SuperRelevant = 1;
			Something += 2.0f;
		}
		else
			SuperRelevant = Actor->bAlwaysRelevant || (NetOwner->ViewTarget && (Actor->IsOwnedBy(NetOwner->ViewTarget) || Actor->Instigator == NetOwner->ViewTarget));

		FLOAT Pri = Actor->GetNetPriority( Sent, Time, InConnection->BestLag);
		Priority = appRound( (Something + 3.0f) * Pri * 65536.f);

		if ( Actor->bNetOptional )
			Priority -= 3000000;
	}
	friend INT Compare( const FActorPriority* A, const FActorPriority* B )
	{
		return B->Priority - A->Priority;
	}
};

//Fix this if old relevancy was toggled
//If True, actor is to be globally checked
inline FLOAT UpdateNetTag( AActor* Other, FLOAT DeltaTime)
{
	Other->NetTag &= 0x7FFFFFFF; //Become positive
	FLOAT* NetTag = (FLOAT*) &Other->NetTag;
	if ( *NetTag > 120.f || appIsNan(*NetTag) )
		*NetTag = appFrand() * 99.f;
	else if ( *NetTag > 99.f )
		*NetTag -= 99.f;
	FLOAT TestVal = *NetTag;
	*NetTag += DeltaTime;
	return TestVal;
}


//See if actor is relevant to a player
static UBOOL ActorIsRelevant( AActor* Actor, APlayerPawn* Player, AActor* Viewer, FVector4* ViewPos, const FVector4& EndOffset)
{
//	if ( Actor->bAlwaysRelevant ) //Handled by super relevant
//		return 1;
	if ( Actor->AmbientSound && ((*ViewPos-Viewer->Location).SizeSquared() < Square(25.0f*Actor->SoundRadius)) )
		return 1;
	if ( Actor->Owner )
	{
		if ( Actor->IsOwnedBy(Viewer) )
			return 1;
		if ( Actor->Owner->bIsPawn && ((APawn*)Actor->Owner)->Weapon == Actor )
			return ActorIsRelevant( Actor->Owner, Player, Viewer, ViewPos, EndOffset);
	}
	if ( (Actor->bHidden || Actor->bOnlyOwnerSee) && !Actor->bBlockPlayers && !Actor->AmbientSound )
		return 0;
	ST_Traces++;
#if ASM
	FVector4 Dest;
	__asm
	{
		mov      eax,[Actor]
		mov      esi,[EndOffset]
		lea      ecx,[Dest]
		movaps   xmm0,[esi]			//x0: EndOffset
		add      eax,208			//Move towards Actor->Location
		movups   xmm2,[eax]			//x2: Location
		add      eax,192			//Move towards Actor->CollisionRadius
		movlps   xmm3,[eax]			//x3: |CR|CH|..|..|
		shufps   xmm3,xmm3,16		//x3: |CR|CR|CH|CR| >>> 0b00010000
		mulps    xmm0,xmm3			//x0: Cylinder offset
		addps    xmm0,xmm2			//x0: Location + Cylinder offset
		movaps   [ecx],xmm0			//Store in Dest
	}
	return ((UXC_Model*) Viewer->XLevel->Model)->FastRepLineCheck( ViewPos, &Dest);
#elif ASMLINUX
	GCC_FVector4_ref(Dest); //GCC isn't aligning FVector4 locals
	__asm__ __volatile__(
	"movups    (%%esi),%%xmm0 \n" //x0: EndOffset |||| THIS IS LIKELY UNALIGNED!!!
	"addl         $208,%%eax  \n" //Move towards Actor->Location
	"movups    (%%eax),%%xmm2 \n" //x2: Location
	"addl         $192,%%eax  \n" //Move towards Actor->CollisionRadius
	"movlps    (%%eax),%%xmm3 \n" //x3: |CR|CH|..|..|
	"shufps $16,%%xmm3,%%xmm3 \n" //x3: |CR|CR|CH|CR| >>> 0b00010000
	"mulps      %%xmm3,%%xmm0 \n" //x0: Cylinder offset
	"addps      %%xmm2,%%xmm0 \n" //x0: Location + Cylinder offset
	"movaps     %%xmm0,(%%ecx) \n" //Store in Dest
	: : "a"(Actor), "c"(Dest), "S"(&EndOffset) : "memory" );
	return ((UXC_Model*) Viewer->XLevel->Model)->FastRepLineCheck( ViewPos, Dest);
#else
	FVector4 Dest;
	Dest.X = Actor->Location.X + EndOffset.X * Actor->CollisionRadius;
	Dest.Y = Actor->Location.Y + EndOffset.Y * Actor->CollisionRadius;
	Dest.Z = Actor->Location.Z + EndOffset.Z * Actor->CollisionHeight;
	return ((UXC_Model*) Viewer->XLevel->Model)->FastRepLineCheck( ViewPos, &Dest);
#endif
}

//BROWSE IS CALLED TOWARDS LOCALMAPURL DURING INIT, WHICH THEN CALLS LOADMAP!
UBOOL UXC_GameEngine::Browse( FURL URL, const TMap<FString,FString>* TravelInfo, FString& Error )
{
	//We're initializing UGameEngine, we no longer need this hack
	if ( UXC_GameEngine::StaticClass()->PropertiesSize < UGameEngine::StaticClass()->PropertiesSize )
		Exchange( UGameEngine::StaticClass()->PropertiesSize, UXC_GameEngine::StaticClass()->PropertiesSize );

	UBOOL bDisconnected = URL.HasOption( TEXT("failed")) || URL.HasOption( TEXT("entry"));
	if ( bDisconnected )
	{
		ULevel* MyLevel = (&GLevel)[b451Setup];
		if ( MyLevel )
			MyLevel->SetActorCollision(0);
		UnHookEngine();
	}

	if ( bEnableDebugLogs )
		debugf( NAME_XC_Engine, TEXT("Browse() Start: %s %s %i"), *URL.Protocol, *URL.Host, URL.Port);
	UBOOL result = Super::Browse( URL, TravelInfo, Error);
	if ( bEnableDebugLogs )
		debugf( NAME_XC_Engine, TEXT("Browse() End"));

	if ( bDisconnected || !GetActiveLevel(this) )
	{
		HookNatives( true); //Full GNative hooks
		HookEngine( NULL ); //Forces limited hook of standard stuff
		appMemzero( &PreLoginHooks, 48);
		AdminLoginHook = NULL;
	}

	return result;
}


/////////////////////
// Compressor thread
/////////////////////
struct FAutoCompressorThread
{
public:
	// Variables.
	DWORD			tId;
	UPackageMap*	PackageMap;
	INT				MapCount;
	TCHAR			CompressedExt[12];
#if __UNIX__
	pthread_t		ACThread;
#endif

#if _MSC_VER
	HANDLE			hThread;
#endif

	// Functions.
	FAutoCompressorThread()	{}
	FAutoCompressorThread( UPackageMap* InPackageMap, INT InMapCount=GMapLoadCount);
};



// Priority thread entrypoint.
#if __UNIX__
void* AutoCompressorThreadEntry(void* Arg)
#else
DWORD STDCALL AutoCompressorThreadEntry(void* Arg)
#endif
{
	appSleep( 0.2f);
	FAutoCompressorThread* ACT = (FAutoCompressorThread*)Arg;
	INT LenExt = appStrlen( ACT->CompressedExt);

	INT i = 0;
	UBOOL bLeave = 0;
	while ( !bLeave )
	{
		appSleep( 0.002f);
		if ( GIsRequestingExit) //Game about to exit
			return 0;
		if ( FPlatformAtomics::InterlockedCompareExchange(&GMapLoadCount, 0, 0) != ACT->MapCount) //New map has loaded
			break;
		if ( FPlatformAtomics::InterlockedCompareExchange(&GIsLoadingMap, 0, 0) != 0) //Map is loading!!
		{
			appSleep( 0.1f);
			continue;
		}
		if ( FPlatformAtomics::InterlockedCompareExchange(&GLockPackageMap, 1, 0)) //Other thread is compressing
		{
			appSleep( 0.1f);
			continue;
		}
		if ( i < ACT->PackageMap->List.Num() )
		{
			ULinker* LNKh = (ULinker*)ACT->PackageMap->List(i).Linker;
			INT Len = LNKh->Filename.Len();
			if ( Len < 250 - LenExt ) //Avoid string formatting crash
			{
				TCHAR FileName_Base[256];
				appStrcpy( FileName_Base, *LNKh->Filename);
				TCHAR FileName_LZMA[256];
				appSprintf( FileName_LZMA, TEXT("%s%s"), FileName_Base, ACT->CompressedExt);


				if ( GXFileManager->FileSize( FileName_LZMA) <= 0 )
				{	//Unsafe operations ended, let's unlock main thread's loaders
					FPlatformAtomics::InterlockedExchange(&GLockPackageMap, 0);
					TCHAR Error[256]; Error[0] = 0;
					TCHAR FileName_Temp[256];
					appStrcpy( FileName_Temp, FileName_LZMA);
					FileName_Temp[Len + LenExt - 1] = 't'; //TEMP INDICATOR
					TCHAR CMPMessage[600];
					appSprintf( CMPMessage, TEXT("AutoCompressing %s to %s"), FileName_Base, FileName_LZMA);
					ThreadedLog( NAME_Log, CMPMessage );
					LzmaCompress( FileName_Base, FileName_Temp, Error);
					GXFileManager->Move( FileName_LZMA, FileName_Temp);
				}
			}
		}
		else
			bLeave = 1;
		i++;
		FPlatformAtomics::InterlockedExchange(&GLockPackageMap, 0);
	}
	return 0;
}

FAutoCompressorThread::FAutoCompressorThread(UPackageMap* InPackageMap, INT InMapCount)
	:	PackageMap(InPackageMap)
	,	MapCount(InMapCount)
{
	guard(WTF);
#ifdef _MSC_VER
	appSprintf( CompressedExt, COMPRESSED_EXTENSION);
	FixFilename( CompressedExt); //Init safe File Manager if necessary

	debugf(NAME_Log, TEXT("Starting LZMA compressor...") );


	DWORD exitCode = 0;

	HANDLE hThread = CreateThread( NULL, 0, AutoCompressorThreadEntry, this, 0, &tId);
	debugf(NAME_XC_Engine, TEXT("Thread ID: %i"), tId);

	//Debugging code I want to keep
	check(hThread);
	GetExitCodeThread(hThread, &exitCode);
	ResumeThread(hThread);
	WaitForSingleObject(hThread, 1); //Waste 1 millisecond
	GetExitCodeThread(hThread, &exitCode);
	CloseHandle( hThread);
#endif
	unguard;
}

static UBOOL PackageNeedsDemoFix( const TCHAR* PkgName)
{
	guard( PackageNeedsDemoFix);
	TCHAR AnnouncerText[] = {'A','n','n','o','u','n','c','e','r', 0 };
	TCHAR CreditsText[] = {'C','r','e','d','i','t','s', 0 };
	const TCHAR* DemoPkgs[] = { AnnouncerText, CreditsText};

	for ( INT i=0 ; i<2 ; i++ )
		if ( appStricmp(DemoPkgs[i], PkgName) == 0 )
			return 1;
	return 0;
	unguard;
}

#include "XCL_Private.h"

void UXC_Level::SetActorCollision( UBOOL bCollision )
{
	guard(UXC_Level::SetActorCollision);

	UXC_GameEngine* XCGE = Cast<UXC_GameEngine>(Engine);
	if ( !XCGE || !XCGE->bCollisionHashHook )
	{
		Super::SetActorCollision(bCollision);
		return;
	}
	
	// Init collision if first time through.
	if( bCollision && !Hash )
	{
		// Init gridded octree.
		guard(StartCollision);
		debugf( NAME_XC_Engine, TEXT("Setting up collision octree [Level Hook] for %s"), GetOuter()->GetName() );
		CollisionCacus = new FCollisionCacus(this);
		Hash = CollisionCacus;
		for( INT i=0; i<Actors.Num(); i++ )
			if( Actors(i) && Actors(i)->bCollideActors )
				Hash->AddActor( Actors(i) );
		unguard;
	}
	else if( Hash && !bCollision )
	{
		// Destroy gridded octree, no need to remove actors.
		guard(EndCollision);
		if ( CollisionCacus == Hash )
		{
			CollisionCacus->Exit();
			delete CollisionCacus;
			Hash = NULL;
			CollisionCacus = NULL;
		}
		else
		{
			debugf( NAME_XC_Engine, TEXT("CollisionCacus not equal to %s hash!!!"), GetOuter()->GetName() );
			Super::SetActorCollision(bCollision);
		}
		unguard;
	}

	unguard;
}

void UXC_Level::Destroy()
{
	guard(ULevel::Destroy);

	// Free allocated stuff.
	SetActorCollision(0);

	if( BrushTracker )
	{
		delete BrushTracker;
		BrushTracker = NULL; /* Required because brushes may clean themselves up. */
	}

	ULevelBase::Destroy();
	unguard;
}

//
// Welcome a new player joining this server.
//
void UXC_Level::WelcomePlayer( UNetConnection* Connection, TCHAR* Optional )
{
	guard(ULevel::WelcomePlayer);

	Connection->PackageMap->Copy( Connection->Driver->MasterMap );
	Connection->SendPackageMap();
	Connection->Logf( TEXT("XC_ENGINE VERSION=%i"), ((UXC_GameEngine*)Engine)->Version ); //Find a way to make clients interpret this
	if( Optional[0] )
		Connection->Logf( TEXT("WELCOME LEVEL=%s LONE=%i %s"), GetOuter()->GetName(), GetLevelInfo()->bLonePlayer, Optional );
	else
		Connection->Logf( TEXT("WELCOME LEVEL=%s LONE=%i"), GetOuter()->GetName(), GetLevelInfo()->bLonePlayer );
	Connection->FlushNet();

	unguard;
}

void UXC_Level::TickNetServer( FLOAT DeltaSeconds )
{
	guard(UXC_Level::TickNetServer);

	if ( ! ((UXC_GameEngine*)Engine)->bUseNewRelevancy )
	{
		Super::TickNetServer( DeltaSeconds);
		return;
	}

	ThreadedLogFlush();
	// Update window title
	if( (INT)(TimeSeconds-DeltaSeconds)!=(INT)(TimeSeconds.GetFloat()) )
		debugf( NAME_Title, LocalizeProgress(TEXT("RunningNet"),TEXT("Engine")), *GetLevelInfo()->Title, *URL.Map, NetDriver->ClientConnections.Num() );
	
	clock(NetTickCycles);
	INT Updated = ServerTickClients( DeltaSeconds);
	unclock(NetTickCycles);

	// Stats.
	static INT SkipCount = 0;
	static FLOAT SkipDelta = 0;
	SkipCount++;
	SkipDelta += DeltaSeconds;
	if( Updated && (appFrand() < DeltaSeconds * 20.f) )
	{
		static INT NumActors;
		NumActors = 0;
		FLOAT FrameRate;
		for( INT i=0; i<NetDriver->ClientConnections.Num(); i++ )
		{
			UNetConnection* Connection = NetDriver->ClientConnections(i);
			if( Connection->Actor && Connection->State==USOCK_Open )
			{
				if( Connection->UserFlags&1 )
				{
					if ( !NumActors ) //Prevent multiple UserFlag clients from using extra CPU cycles
					{
						FrameRate = _Reciprocal(SkipDelta) * (FLOAT)SkipCount * GetLevelInfo()->TimeDilation;
						for( INT ii=0; ii<Actors.Num(); ii++ )
							NumActors += Actors(ii)!=NULL;
					}

					FString Stats = FString::Printf
					(
						TEXT("r=%03.1f cli=%i act=%03.1f (%i) net=%03.1f pv/c=%i rep/c=%i trace/c=%i"),
						FrameRate,
						NetDriver->ClientConnections.Num(),
						GSecondsPerCycle * 1000 * ActorTickCycles,
						NumActors,
						GSecondsPerCycle * 1000 * NetTickCycles,
						NumPV/NetDriver->ClientConnections.Num(),
						NumReps/NetDriver->ClientConnections.Num(),
						ST_Traces/(NetDriver->ClientConnections.Num() * SkipCount)
					);
					Connection->Actor->eventClientMessage( *Stats, NAME_None, 0 );

				}
				if( Connection->UserFlags&2 )
				{
					FString Stats = FString::Printf
					(
						TEXT("snd=%02.1f recv=%02.1f"),
						GSecondsPerCycle*1000*Connection->Driver->SendCycles,
						GSecondsPerCycle*1000*Connection->Driver->RecvCycles
					);
					Connection->Actor->eventClientMessage( *Stats, NAME_None, 0 );
				}
			}
		}
		SkipDelta = 0;
		SkipCount = 0;
		ST_Traces = 0;
	}
	NumPV = 0;
	NumReps = 0;

	unguard;
}

//Multithreaded relevancy builder (one of these days)
MS_ALIGN(16) struct FConnectionRelevancyList
{
	UXC_NetConnectionHack* Connection;	//O=0
	FConnectionRelevancyList* Next;	//O=4

	FActorPriority* PriorityArray;	//O=8
	FActorPriority** PriorityRefs;	//O=12

	AActor** ConsiderList;			//O=16
	AActor** OwnedConsiderList;		//O=20
	AActor* Viewer;					//O=24
	INT ConsiderListSize;			//O=28
	INT OwnedConsiderListSize;		//O=32

	FRotator Rotation;				//O=36
	FVector4 Location;				//O=48 (FVector4 has alignment=16)

	FConnectionRelevancyList( UNetConnection* InConn, FConnectionRelevancyList* InNext, INT InID)
		:	Connection( (UXC_NetConnectionHack*)InConn)
		,	Next( InNext)
	{};

	//This is a multithread nightmare, let's put it on main thread for now
	void GetViewerCoordinates()
	{
		// Get viewer coordinates.
		Location  = FVector4( Connection->Actor->Location, -1.f);
		Rotation  = Connection->Actor->ViewRotation;
		Viewer    = Connection->Actor;
		Connection->Actor->eventPlayerCalcView( Viewer, Location, Rotation );
		check(Viewer);
	}
	
	void PawnBandwidthHack( UActorChannel* Channel)
	{
		if ( !Channel->Recent.Num() )
			return;
		guard(PawnBandwidthHack);
		APawn* OldPawn = (APawn*)Channel->Recent.GetData();
		APawn* NewPawn = (APawn*)Channel->Actor;
		FLOAT Delta = Connection->Driver->Time - Channel->LastUpdateTime;
		FVector CalcLocation = OldPawn->Location + OldPawn->Velocity * Delta;
		if ( (CalcLocation - NewPawn->Location).SizeSquared() < 1.0f )
			OldPawn->Location = NewPawn->Location;
		unguard;
	}
	
	void RelevancyLoop( FLOAT RealDelta, FLOAT MaxTickRate, INT ConnID)
	{
		guard(RelevancyLoop);
		INT ConsiderCount = 0;


		guard(SetupTraceStart);
		GetViewerCoordinates();
		if ( Connection->TickCount & 1 ) //Happens every 2 frames
		{
			FLOAT TMult = (Connection->TickCount & 2) ? 0.4f : 0.9f; //Pick 0.4 or 0.9 seconds (one frame 0.4, next 0.9 and so on)
			FVector4 Ahead = Location + Viewer->Velocity * TMult;
			if ( Viewer->Base ) //Add platform/lift velocity to the Ahead calc
				Ahead += Viewer->Base->Velocity * TMult;
			if ( ((UXC_Model*)Viewer->XLevel->Model)->FastRepLineCheck( &Ahead, &Location) )
				Location = Ahead;
		}
		unguard;

		TArray<AActor*>* ST = Connection->GetSentTemporaries();
		INT i;

		//Skip sent temporaries (set NetTag to negative)
		for ( i=0 ; i<ST->Num() ; i++ )
			(*ST)(i)->NetTag |= 0x80000000;

		FVector4 Dir( Rotation.Vector(), 0.f); //0.f is critical
		TMap<AActor*,UActorChannel*>* AC = Connection->GetActorChannels();
		TArray<UChannel*>* OC = Connection->GetOpenChannels();
		
		//Ez mode
		FLOAT OffsetFloat = (FLOAT)ConnID * 0.045;
		UBOOL bChannelsSaturated = OC->Num() > 800;

		//Prioritize owned actors first
		guard(PrioritizeOwned);
		for ( i=0 ; i<OwnedConsiderListSize ; i++ )
		{
			AActor* Actor = OwnedConsiderList[i];
			if ( !(Actor->NetTag & 0x80000000) )
			{
				UBOOL bPrioritize = 0;
				Actor->NetTag |= 0x80000000;
				UActorChannel* Channel = AC->FindRef(Actor);
				if ( Actor->bAlwaysRelevant ) //Already passed general check time earlier
				{
					bPrioritize = !( Actor->CheckRecentChanges() && Channel && Channel->Recent.Num() && !Channel->Dirty.Num() && Actor->NoVariablesToReplicate( (AActor*) &Channel->Recent(0)) );
					if ( !bPrioritize ) //bAlwaysRelevant actor skips an update
						Channel->RelevantTime = Connection->Driver->Time;
				}
				else
				{
					FLOAT UF = Actor->UpdateFrequency( Viewer, Dir, Location);
					bPrioritize = (UF >= MaxTickRate) || (appRound( (*(FLOAT*)&Actor->NetTag + OffsetFloat)*UF) != appRound( (*(FLOAT*)&Actor->NetTag + OffsetFloat - RealDelta)* UF));
				}
				if ( bPrioritize )
				{
					PriorityArray[ConsiderCount] = FActorPriority( EPri_Owned, Location, Dir, Connection, Channel, Actor);
					PriorityRefs[ConsiderCount] = PriorityArray + ConsiderCount;
					if ( bChannelsSaturated && !Actor->bHidden && !Actor->Inventory && (Actor != Connection->Actor) )
						PriorityRefs[ConsiderCount]->SuperRelevant = 0; //We can afford to lose visible owned actors
					ConsiderCount++;
				}
			}
		}
		unguardf( (TEXT("[%i/%i]"), i, OwnedConsiderListSize));

		//Spectator: artificially reduce tickrate to half when updating non-owned stuff
		//Update owned stuff as usual, but update others at half the rate
		UBOOL bSkipMainPri = false; //Do not prioritize, simply fix NetTag
		if ( Connection->Actor->PlayerReplicationInfo && Connection->Actor->PlayerReplicationInfo->bIsSpectator )
		{
			if ( Connection->TickCount & 1 == 0 ) //This will prevent the general loop
			{
				bSkipMainPri = true;
				for ( i=0 ; i<OwnedConsiderListSize ; i++ )
					OwnedConsiderList[i]->NetTag &= 0x7FFFFFFF;
				for ( i=0 ; i<ST->Num() ; i++ )
					(*ST)(i)->NetTag &= 0x7FFFFFFF;
			}
			RealDelta *= 2.f;
			MaxTickRate *= 0.5f;
		}

		//Actors exist, have role
		if ( !bSkipMainPri )
		{
			guard(PrioritizeActors);
			for ( i=0 ; i<ConsiderListSize ; i++ )
			{
				AActor* Actor = ConsiderList[i];
				if ( Actor->NetTag & 0x80000000 ) //NetTag is negative, actor not prioritized here
					Actor->NetTag &= 0x7FFFFFFF;
				else
				{
					UBOOL bPrioritize = 0;
					UActorChannel* Channel = AC->FindRef(Actor);
					if ( Actor->bAlwaysRelevant ) //Already passed general check time earlier
					{
						bPrioritize = !( Actor->CheckRecentChanges() && Channel && Channel->Recent.Num() && !Channel->Dirty.Num() && Actor->NoVariablesToReplicate( (AActor*) &Channel->Recent(0)) );
						if ( !bPrioritize ) //bAlwaysRelevant actor skips an update
							Channel->RelevantTime = Connection->Driver->Time;
					}
					else
					{
						FLOAT UF = Actor->UpdateFrequency( Viewer, Dir, Location);
						bPrioritize = (UF >= MaxTickRate) || (appRound( (*(FLOAT*)&Actor->NetTag + OffsetFloat)*UF) != appRound( (*(FLOAT*)&Actor->NetTag + OffsetFloat - RealDelta)* UF));
					}
					if ( bPrioritize )
					{
						PriorityArray[ConsiderCount] = FActorPriority( EPri_General, Location, Dir, Connection, Channel, Actor);
						PriorityRefs[ConsiderCount] = PriorityArray + ConsiderCount;
						if ( bChannelsSaturated && !Actor->bHidden && (Actor != Connection->Actor) )
							PriorityRefs[ConsiderCount]->SuperRelevant = 0; //We can afford to lose visible actors
						ConsiderCount++;
					}
				}
			}
			unguardf( (TEXT("[%i/%i]"), i, ConsiderListSize));
		}

		Connection->LastRepTime = Connection->Driver->Time;

		//Prioritize
		for ( i=1 ; i<ConsiderCount ; i++ )
			if ( PriorityRefs[i]->Priority > PriorityRefs[i-1]->Priority )
			{
				Exchange( PriorityRefs[i], PriorityRefs[i-1] );
				if ( i == 1 )
					continue;
				i -= 2;
			}


		INT Updated = 0;
		FVector4 EndOffset( VRand(), 1.f);
		EndOffset.X *= 0.95;
		EndOffset.Y *= 0.95;
		EndOffset.Z *= 0.95;
		UNetDriver* NetDriver = Connection->Driver;

		guard(Relevancy);
		for ( i=0 ; i<ConsiderCount ; i++ )
		{
			UActorChannel* Channel = PriorityRefs[i]->Channel;
			AActor* Actor = PriorityRefs[i]->Actor;
			UBOOL IsRelevant = PriorityRefs[i]->SuperRelevant;
			if ( IsRelevant == -1 ) //Do not run relevancy checks
				IsRelevant = 0;
			else if ( !IsRelevant && (!Channel || (NetDriver->Time - Channel->RelevantTime > 0.3f)) )
				IsRelevant = ActorIsRelevant( Actor, Connection->Actor, Viewer, &Location, EndOffset);

			if( IsRelevant || (Channel && NetDriver->Time-Channel->RelevantTime < NetDriver->RelevantTimeout) )
			{
				Actor->XLevel->NumPV++;
				if( !Channel && Connection->PackageMap->ObjectToIndex(Actor->GetClass())!=INDEX_NONE )
				{
					// Create a new channel for this actor.
					Channel = (UActorChannel*)Connection->CreateChannel( CHTYPE_Actor, 1, INDEX_NONE);
					if( Channel )
						Channel->SetChannelActor( Actor );
				}
				if ( Channel )
				{
					if ( !Connection->IsNetReady(0) )
						break;
					if( IsRelevant )
						Channel->RelevantTime = NetDriver->Time + 0.1 + appFrand();
					if( Channel->IsNetReady(0) )
					{
						//Hack for simulated pawns, reduces bandwidth
						if ( Actor->bSimulatedPawn && (Actor->Physics != PHYS_None) )
							PawnBandwidthHack( Channel);
						Channel->ReplicateActor();
						Updated++;
					}
					if ( !Connection->IsNetReady(0) )
						break;
				}
			}
			else if ( Channel && (Actor != Connection->Actor)) //Never close the local player's channel
				Channel->Close();
		}
		unguardf( (TEXT("[%i/%i]"), i, ConsiderCount));
		unguard;
	}


} GCC_ALIGN(16);


INT UXC_Level::ServerTickClients( FLOAT DeltaSeconds )
{
	guard(UXC_Level::ServerTickClients);

	DeltaSeconds *= _Reciprocal( GetLevelInfo()->TimeDilation);

	INT i;
	INT Updated=0;
	FMemMark Mark(GMem);
	FConnectionRelevancyList* ConnList = NULL;

	guard( MakeConnectionLists);
	for( i=NetDriver->ClientConnections.Num()-1; i>=0; i-- ) //Reverse order, so chained list isn't flipped
	{
		UNetConnection* Conn = NetDriver->ClientConnections(i);
		check( Conn);
		check( Conn->State==USOCK_Pending || Conn->State==USOCK_Open || Conn->State==USOCK_Closed);

		if ( !Conn->Actor || !Conn->IsNetReady(0) || Conn->State!=USOCK_Open )
			continue;
		if ( Conn->Driver->Time - Conn->LastReceiveTime > 1.5f )
			continue;
		ConnList = new(GMem,1,16) FConnectionRelevancyList( Conn, ConnList, Updated);
		Updated++;
	}
	//No clients, abort here
	if ( !Updated )
		return 0;
	unguard;

	AActor** ConsiderList;
	INT ConsiderListSize = 0;

	
	guard( MakeConsiderList);
	INT ActorListSize = Actors.Num() + 2 + 1 - iFirstNetRelevantActor; //That +1 is a buffer!!!
	INT OwnedListPos = ActorListSize;
	ConsiderList = new(GMem, ActorListSize+1) AActor*;
	// Add LevelInfo
	if( Actors(0) && (Actors(0)->RemoteRole!=ROLE_None) )
	{
		FLOAT OldTime = UpdateNetTag( Actors(0), DeltaSeconds);
		if ( appRound( OldTime*Actors(0)->NetUpdateFrequency) != appRound( *((FLOAT*)&Actors(0)->NetTag)*Actors(0)->NetUpdateFrequency) )
		{
			Actors(0)->bAlwaysRelevant = true; //Important to enforce this
			*(FLOAT*) &Actors(0)->NetTag += appFrand() * 0.05; //Randomizer
			ConsiderList[ConsiderListSize++] = Actors(0);
		}
	}
	//Non-owned actors from (0 to max), owned actors from (max to 0)
	FLOAT ArOffset = 0.f; //bAlwaysRelevant actors spreader
	for( i=iFirstNetRelevantActor; i<Actors.Num(); i++ ) //Consider all actors for now
	{
		AActor* Actor = Actors(i);
		if ( Actor && Actor->RemoteRole != ROLE_None )
		{
			//NetTag rules: always positive, between 0 and 99 (up to 120)
			FLOAT OldTime = UpdateNetTag( Actor, DeltaSeconds);
			UBOOL bPass = 0;
			if ( Actor->bAlwaysRelevant )
			{
				//bAlwaysRelevant actors are checked at the same time
				if ( appRound( (OldTime+ArOffset)*Actor->NetUpdateFrequency) != appRound( (*((FLOAT*)&Actor->NetTag)+ArOffset)*Actor->NetUpdateFrequency) )
				{
					*(FLOAT*) &Actor->NetTag += appFrand() * 0.05; //Randomizer (is it necessary?)
					ArOffset += 0.023f;
					bPass = 1;
				}
			}
			else if ( !(Actor->bNetOptional && (Actor->GetClass()->GetDefaultActor()->LifeSpan - 0.15f >= Actor->LifeSpan)) )
				bPass = 1;
			
			if ( bPass )
			{
				if ( Actor->Owner || Actor->Instigator )
					ConsiderList[--OwnedListPos] = Actor;
				else
					ConsiderList[ConsiderListSize++] = Actor;
			}
		}
	}
	//Sort owned list by ConnectionID and move to end of ConsiderList incrementally
	for ( FConnectionRelevancyList* ConnLink=ConnList ; ConnLink ; ConnLink=ConnLink->Next )
	{
		ConnLink->OwnedConsiderList = &ConsiderList[ConsiderListSize];
		ConnLink->OwnedConsiderListSize = 0;
		for ( i=OwnedListPos ; i<ActorListSize ; i++ )
		{
			APlayerPawn* NetOwner = ConnLink->Connection->Actor;
			//Prioritize instigator 100%, Pawns are always self-instigated (hopefully)
			if ( ConsiderList[i]->Instigator == NetOwner || (!ConsiderList[i]->Instigator && ConsiderList[i]->IsOwnedBy(NetOwner)) )
			{
				ConsiderList[ConsiderListSize++] = ConsiderList[i];
				ConsiderList[i] = ConsiderList[OwnedListPos++];
				ConnLink->OwnedConsiderListSize++;
			}
		}
	}
	//Move any leftovers not belonging to connected players (filter them out?, useful for bot games)
	while ( OwnedListPos < ActorListSize )
		ConsiderList[ConsiderListSize++] = ConsiderList[OwnedListPos++];
	unguard;


	guard( ArrayProcess);

	FLOAT MaxTickRate = Engine->GetMaxTickRate(); //This should become an argument
	if ( MaxTickRate == 0 && (DeltaSeconds != 0.f) )
		MaxTickRate = _Reciprocal( DeltaSeconds);
	//Allocate general chunks, they will be reused by each connection
	FActorPriority* GPriorityArray = new(GMem,ConsiderListSize) FActorPriority;
	FActorPriority** GPriorityRefs = new(GMem,ConsiderListSize) FActorPriority*;
	i = 0;
	for ( FConnectionRelevancyList* ConnLink=ConnList ; ConnLink ; ConnLink=ConnLink->Next )
	{
		ConnLink->PriorityArray = GPriorityArray;
		ConnLink->PriorityRefs = GPriorityRefs;
		ConnLink->ConsiderList = ConsiderList;
		ConnLink->ConsiderListSize = ConsiderListSize;
		ConnLink->Connection->TickCount++;
		NetTag++;
		ConnLink->RelevancyLoop( DeltaSeconds, MaxTickRate, i);
		i++;
	}
	unguardf( (TEXT("[%i/%i]"), i, Updated));

	Mark.Pop();
	return Updated;
	unguard;
}

//This doesn't appear to be working, dammit
//Prevents heavy log spamming
//TODO: Figure out a way to allow a client to rebroadcast a game?
EAcceptConnection UXC_Level::NotifyAcceptingConnection()
{
	check(NetDriver);
	static FTime LastNAC = 0.f; //Don't initialize a static variable using appSeconds(), just in case
	FTime CurTime = appSeconds();
	if( NetDriver->ServerConnection )
	{
		// We are a client and we don't welcome incoming connections.
		// Don't fill a client's log
		if ( CurTime - LastNAC > 1.f )
		{
			LastNAC = CurTime;
			debugf( NAME_DevNet, TEXT("NotifyAcceptingConnection: Client refused") );
		}
		return ACCEPTC_Reject;
	}
	else if( GetLevelInfo()->NextURL!=TEXT("") )
	{
		// Server is switching levels.
		// 100% useless, populated v436 servers suffer a lot from this
//		debugf( NAME_DevNet, TEXT("NotifyAcceptingConnection: Server %s refused"), GetName() );
		return ACCEPTC_Ignore;
	}
	else
	{
		// Server is up and running.
		// Log everything during the first second of the level
		// Otherwise add a 0.25s interval to avoid DDoS's from filling log
		if ( (NetDriver->Time.GetFloat() < 1.f) || (CurTime - LastNAC > 0.25f) )
		{
			LastNAC = CurTime;
			debugf( NAME_DevNet, TEXT("NotifyAcceptingConnection: Server %s accept"), GetName() );
		}
		return ACCEPTC_Accept;
	}
}

void UXC_Level::NotifyAcceptedConnection( class UNetConnection* Connection )
{
	Super::NotifyAcceptedConnection( Connection);
}


//====================================================
//====================================================
// Brush tracker fixer
//====================================================
//====================================================

#ifndef DISABLE_ADDONS

UBOOL FXC_BrushTrackerFixer::Init()
{
	guard( FXC_BrushTrackerFixer::Init);
	check( XLevel);

	if ( XLevel->BrushTracker || !XLevel->Actors(0) )
		return false;

	debugf( NAME_XC_Engine, TEXT("Checking for Moving Brush Tracker insertion..."));
	for( INT i=0; i<XLevel->Actors.Num(); i++ )
	{
		guard(MoverCheck);
		if( StaticMover( XLevel->Actors(i)) ) //Static mover
		{
			if ( !XLevel->BrushTracker )
				XLevel->BrushTracker = GNewBrushTracker( XLevel );
			XLevel->BrushTracker->Flush( XLevel->Actors(i) );
			XLevel->BrushTracker->Update( XLevel->Actors(i) );
			StaticMovers.AddItem( (AMover*) XLevel->Actors(i) );
		}
		unguard;
	}
	if ( !StaticMovers.Num() )
		return false;
	StaticMovers.Shrink();
	return true;

	unguard;
}

void FXC_BrushTrackerFixer::Exit()
{
	if ( StaticMovers.Num() > 0 )
		StaticMovers.Empty();
}

UBOOL FXC_BrushTrackerFixer::IsTyped( const TCHAR* Type)
{
	return appStricmp( Type, TEXT("BrushTrackerFixer")) == 0;
}

INT FXC_BrushTrackerFixer::Tick( FLOAT DeltaSeconds)
{
	guard( FXC_BrushTrackerFixer::Tick)
	if ( !XLevel || StaticMovers.Num() <= 0 || !XLevel->BrushTracker )
		return 0; //Should deinitialize
	if ( XLevel->Actors.Num() <= 0 || !XLevel->Actors(0) || !XLevel->GetLevelInfo()->bBegunPlay )
		return 1; //Should not deinitialize
	for( INT i=0; i<StaticMovers.Num(); i++ )
	{
		if ( !StaticMovers(i) || StaticMovers(i)->bDeleteMe || StaticMovers(i)->bHidden )
			StaticMovers.Remove(i);
		else
			XLevel->BrushTracker->Update( StaticMovers(i) );
	}
	//TestLevel->BrushTracker->Flush( TestLevel->Actors(i) );

	return 1;
	unguard;
}

#endif